#!/usr/bin/env python
"""
Python helper script for PostScript import

Copyright (C) 2012 ~suv, <suv-sf@users.sourceforge.net>

'default_import()' is based on Inkscape's 'ps2pdf-ext.py':
Copyright (C) 2008 Stephen Silver

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

# standard library
import re
import sys
# from run_command.py
import os
import tempfile

# local library
import inkex
#from run_command import run

try:
    inkex.localize()
except:
    import gettext
    _ = gettext.gettext

#DEBUG = True

# all papersizes known to Ghostscript
# <http://www.ghostscript.com/doc/9.06/Use.htm#Known_paper_sizes>
gs_papersizes = {
        '11x17'       : (792, 1224),
        'ledger'      : (1224, 792),
        'legal'       : (612, 1008),
        'letter'      : (612, 792),
        'lettersmall' : (612, 792),
        'archE'       : (2592, 3456),
        'archD'       : (1728, 2592),
        'archC'       : (1296, 1728),
        'archB'       : (864, 1296),
        'archA'       : (648, 864),
        'a0'          : (2384, 3370),
        'a1'          : (1684, 2384),
        'a2'          : (1191, 1684),
        'a3'          : (842, 1191),
        'a4'          : (595, 842),
        'a4small'     : (595, 842),
        'a5'          : (420, 595),
        'a6'          : (297, 420),
        'a7'          : (210, 297),
        'a8'          : (148, 210),
        'a9'          : (105, 148),
        'a10'         : (73, 105),
        'isob0'       : (2835, 4008),
        'isob1'       : (2004, 2835),
        'isob2'       : (1417, 2004),
        'isob3'       : (1001, 1417),
        'isob4'       : (709, 1001),
        'isob5'       : (499, 709),
        'isob6'       : (354, 499),
        'c0'          : (2599, 3677),
        'c1'          : (1837, 2599),
        'c2'          : (1298, 1837),
        'c3'          : (918, 1298),
        'c4'          : (649, 918),
        'c5'          : (459, 649),
        'c6'          : (323, 459),
        'flsa'        : (612, 936),
        'flse'        : (612, 936),
        'halfletter'  : (396, 612),
        'hagaki'      : (283, 420)
        }

def run(command_format, prog_name, verbose):
    """
    Run a command that generates an SVG (or PDF) file from an input file.
    On success, outputs the contents of the resulting file to stdout, and exits
    with a return code of 0.
    On failure, outputs an error message to stderr, and exits with a return
    code of 1.
    """
    svgfile = tempfile.mktemp(".svg")
    if verbose:
        inkex.debug(command_format % svgfile)
    command = command_format % svgfile
    msg = None
    # ps2pdf may attempt to write to the current directory, which may not
    # be writeable, so we switch to the temp directory first.
    try:
        os.chdir(tempfile.gettempdir())
    except Exception:
        pass
    # In order to get a return code from the process, we use subprocess.Popen
    # if it's available (Python 2.4 onwards) and otherwise use popen2.Popen3
    # (Unix only).  As the Inkscape package for Windows includes Python 2.5,
    # this should cover all supported platforms.
    try:
        try:
            from subprocess import Popen, PIPE
            p = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
            rc = p.wait()
            out = p.stdout.read()
            err = p.stderr.read()
        except ImportError:
            try:
                from popen2 import Popen3
                p = Popen3(command, True)
                p.wait()
                rc = p.poll()
                out = p.fromchild.read()
                err = p.childerr.read()
            except ImportError:
                # shouldn't happen...
                msg = "Neither subprocess.Popen nor popen2.Popen3 is available"
        if rc and msg is None:
            msg = "%s failed:\n%s\n%s\n" % (prog_name, out, err)
    except Exception, inst:
        msg = "Error attempting to run %s: %s" % (prog_name, str(inst))

    # If successful, copy the output file to stdout.
    if msg is None:
        if os.name == 'nt':  # make stdout work in binary on Windows
            import msvcrt
            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
        try:
            f = open(svgfile, "rb")
            data = f.read()
            sys.stdout.write(data)
            f.close()
        except IOError, inst:
            msg = "Error reading temporary file: %s" % str(inst)

    # Clean up.
    try:
        os.remove(svgfile)
    except Exception:
        pass

    # Ouput error message (if any) and exit.
    if msg is not None:
        sys.stderr.write(msg + "\n")
        sys.exit(1)
    else:
        sys.exit(0)

class InputEPS(inkex.Effect):
    def __init__(self):
        # Call base class construtor.
        inkex.Effect.__init__(self)
        self.document=[]
        self.input_bbox=[]
        self.input_pages=[]
        self.input_papersizes=[]
        self.input_page_bbox=[]
        self.verbose = False

        # Define option
        # hidden parameter
        self.OptionParser.add_option("--dEPSCrop",
            action="store", type="inkbool",
            dest="dEPSCrop", default=True,
            help="")
        # presets
        self.OptionParser.add_option("--papersize",
            action="store", type="string",
            dest="papersize", default="EPSCrop",
            help="")
        self.OptionParser.add_option("--orientation",
            action="store", type="string",
            dest="orientation", default="portrait",
            help="")
        # custom options - gs
        # Page parameters
        self.OptionParser.add_option("--gs_sPAPERSIZE_opt",
            action="store", type="inkbool",
            dest="gs_sPAPERSIZE_opt", default=True,
            help="")
        self.OptionParser.add_option("--gs_sPAPERSIZE_val",
            action="store", type="string",
            dest="gs_sPAPERSIZE_val", default="a4",
            help="")
        self.OptionParser.add_option("--gs_sPAPERSIZE_orient",
            action="store", type="string",
            dest="gs_sPAPERSIZE_orient", default="portrait",
            help="")
        self.OptionParser.add_option("--gs_dDEVICEPOINTS_opt",
            action="store", type="inkbool",
            dest="gs_dDEVICEPOINTS_opt", default=False,
            help="")
        self.OptionParser.add_option("--gs_dDEVICEWIDTHPOINTS_val",
            action="store", type="int",
            dest="gs_dDEVICEWIDTHPOINTS_val", default=595,
            help="")
        self.OptionParser.add_option("--gs_dDEVICEHEIGHTPOINTS_val",
            action="store", type="int",
            dest="gs_dDEVICEHEIGHTPOINTS_val", default=842,
            help="")
        self.OptionParser.add_option("--gs_translate_xy_opt",
            action="store", type="inkbool",
            dest="gs_translate_xy_opt", default=False,
            help="")
        self.OptionParser.add_option("--gs_translate_x_val",
            action="store", type="int",
            dest="gs_translate_x_val", default=0,
            help="")
        self.OptionParser.add_option("--gs_translate_y_val",
            action="store", type="int",
            dest="gs_translate_y_val", default=0,
            help="")
        self.OptionParser.add_option("--gs_BOUNDINGBOX_opt",
            action="store", type="inkbool",
            dest="gs_BOUNDINGBOX_opt", default=False,
            help="")
        self.OptionParser.add_option("--gs_dFIXEDMEDIA_opt",
            action="store", type="inkbool",
            dest="gs_dFIXEDMEDIA_opt", default=True,
            help="")
        self.OptionParser.add_option("--gs_dORIENT1_swap",
            action="store", type="inkbool",
            dest="gs_dORIENT1_swap", default=False,
            help="")
        # EPS parameters
        self.OptionParser.add_option("--gs_dEPSbbox_opt",
            action="store", type="string",
            dest="gs_dEPSbbox_opt", default="0",
            help="")
        self.OptionParser.add_option("--gs_dNOEPS_opt",
            action="store", type="inkbool",
            dest="gs_dNOEPS_opt", default=False,
            help="")
        # unsupported tools
        self.OptionParser.add_option("--unsupported_cmd",
            action="store", type="string",
            dest="unsupported_cmd", default="pstopdf",
            help="")
        # tabs
        self.OptionParser.add_option("--tab",
            action="store", type="string",
            dest="tab")
        # global
        self.OptionParser.add_option("--verbose",
            action="store", type="inkbool",
            dest="verbose", default=False,
            help="")

    def effect(sef):
        pass

    def parse(self,file=None):
        """
        Parse document in specified file or on stdin
        """
        try:
            try:
                stream = open(file,'r')
            except:
                stream = open(self.input_file,'r')
        except:
            stream = sys.stdin
        self.document = stream.readlines()
        stream.close()

    def get_sizes(self, infile):
        """
        Collect size information from input PostScript file
        """
        for line in infile:
            if "%%Pages" in line:
                self.input_pages.append(line.strip().split()[1:])
            elif "%%BoundingBox" in line:
                self.input_bbox.append(line.strip().split()[1:])
            elif "%%DocumentPaperSizes" in line:
                self.input_papersizes.append(line.strip().split()[1:])
            elif "%%PageBoundingBox" in line:
                self.input_page_bbox.append(line.strip().split()[1:])

    def report_sizes(self):
        """
        Print various sizes to stderr
        """
        if self.options.dEPSCrop:
            inkex.debug("EPS file: %s" % (self.input_file))
        else:
            inkex.debug("PS file: %s" % (self.input_file))

        if len(self.input_bbox) > 0:
            inkex.debug("Number of bbox entries: %s" % len(self.input_bbox))
            c = 0
            for bbox_entry in self.input_bbox:
                c += 1
                if len(bbox_entry) == 4:
                    inkex.debug("BoundingBox(%s) (x, y): %s, %s"
                            % (c, bbox_entry[0], bbox_entry[1]))
                    inkex.debug("BoundingBox(%s) (w, h): %s, %s"
                            % (c, bbox_entry[2], bbox_entry[3]))
                else:
                    inkex.debug("BoundingBox(%s): %s"
                            % (c, bbox_entry))
        if len(self.input_page_bbox) > 0:
            inkex.debug("Number of pbbox entries: %s" % len(self.input_page_bbox))
            c = 0
            for pbbox_entry in self.input_page_bbox:
                c += 1
                if len(pbbox_entry) == 4:
                    inkex.debug("PageBoundingBox(%s) (x, y): %s, %s"
                            % (c, pbbox_entry[0], pbbox_entry[1]))
                    inkex.debug("PageBoundingBox(%s) (w, h): %s, %s"
                            % (c, pbbox_entry[2], pbbox_entry[3]))
                else:
                    inkex.debug("PageBoundingBox(%s): %s"
                            % (c, pbbox_entry))
        if len(self.input_pages) > 0:
            inkex.debug("Number of pages entries: %s" % len(self.input_pages))
            c = 0
            for pages_entry in self.input_pages:
                c += 1
                inkex.debug("Pages(%s): %s" 
                        % (c, pages_entry[0]))
        if len(self.input_papersizes) > 0:
            inkex.debug("Number of papersizes entries: %s" % len(self.input_papersizes))
            c = 0
            for papersizes_entry in self.input_papersizes:
                c += 1
                inkex.debug("DocumentPaperSizes(%s): %s"
                        % (c, papersizes_entry[0]))
        inkex.debug("------------------------------------------")

    def parse_bbox(self,bbox):
        """
        Parse BoundingBox, relocate at (0,0), return list of w,h,tx,ty
        """
        x0 = int(bbox[0])
        y0 = int(bbox[1])
        x1 = int(bbox[2])
        y1 = int(bbox[3])
        # return new page width, height, translate x, translate y to position at (0,0)
        return((x1-x0), (y1-y0), -x0, -y0)

    def default_import(self, infile):
        """
        Default import via ps2pdf, using -dEPSCrop
        """
        cmd = 'ps2pdf'
        if self.options.dEPSCrop:
            cmd += ' -dEPSCrop'
        # run external command
        if self.verbose:
            inkex.debug(((cmd+' "%s" "%%s"')
                % infile.replace("%","%%"), "ps2pdf", self.verbose))
        run((cmd+' "%s" "%%s"') 
                % infile.replace("%","%%"), "ps2pdf", self.verbose)

    def custom_import(self, infile, cmd0, cmd1, cmd2, cmd3):
        """
        Custom import via gs, using custom options
        """
        cmd_str1 = cmd0
        cmd_str2 = cmd1+cmd2+cmd3
        # run external command
        if self.verbose:
            inkex.debug(((cmd_str1 + ' -sOutputFile="%%s"' + cmd_str2 + ' -f "%s"')
                % infile.replace("%","%%"), "gs", self.verbose))
        run((cmd_str1 + ' -sOutputFile="%%s"' + cmd_str2 + ' -f "%s"')
                % infile.replace("%","%%"), "gs", self.verbose)

    def affect(self, args=sys.argv[1:], output=False, input=True):
        """
        Affect input document with a callback effect
        """
        self.input_file = args[-1]
        self.getoptions(args)
        self.verbose = self.options.verbose
        self.parse()
        self.get_sizes(self.document)
        if self.verbose: self.report_sizes()
        self.effect()
        if output: self.output()
        if input: self.input()

    def input(self):
        """
        Parse input options, construct and run external command string
        """
        infile = self.input_file

        if self.options.tab == '"default_tab"':

            # use ps2pdf, crop EPS
            self.default_import(infile)

        elif self.options.tab == '"papersize_presets_tab"':

            # command string
            cmd0  = 'gs -q -dSAFER -dNOPAUSE -dBATCH' # -sOutputFile=file.pdf

            # options
            cmd1  = ' -sDEVICE=pdfwrite'              # [more options]
            cmd2  = ''                                # postscript code
            cmd3  = ' -c .setpdfwrite'                # before "-f source1.ps [more files]"

            # parse and construct argument string for gs
            if self.options.papersize == "system":
                cmd1 += ''
            elif self.options.papersize == "EPSCrop":
                cmd1 += ' -dEPSCrop'
            elif self.options.papersize == "EPSFitPage":
                cmd1 += ' -dEPSFitPage'
            else: # PAPERSIZE
                try:
                    # test for known PAPERSIZE
                    known_papersize = gs_papersizes[self.options.papersize]
                    # orientation
                    if self.options.orientation == "portrait":
                        cmd1 += ( ' -dDEVICEWIDTHPOINTS=%s -dDEVICEHEIGHTPOINTS=%s'
                                % (known_papersize[0], known_papersize[1]) )
                    else: #landscape
                        cmd1 += ( ' -dDEVICEWIDTHPOINTS=%s -dDEVICEHEIGHTPOINTS=%s'
                                % (known_papersize[1], known_papersize[0]) )
                    cmd1 += " -dFIXEDMEDIA"
                except:
                    # papersize not found in dict
                    # fall back to system default page
                    inkex.errormsg("PAPERSIZE not known to ghostscript, using system default.")
                    cmd1 += ''

            # run external command
            self.custom_import(infile, cmd0, cmd1, cmd2, cmd3)

        elif self.options.tab == '"gs_tab"':

            # command string
            cmd0  = 'gs -q -dSAFER -dNOPAUSE -dBATCH' # -sOutputFile=file.pdf

            # options
            cmd1  = ' -sDEVICE=pdfwrite'              # [more options]
            cmd2  = ''                                # postscript code
            cmd3  = ' -c .setpdfwrite'                # before "-f source1.ps [more files]"

            # Page parameters
            if  self.options.gs_BOUNDINGBOX_opt:
                # try BoundingBox
                if len(self.input_bbox) > 0:
                    if len(self.input_bbox[0]) == 4:
                        # first encountered BoundingBox is valid
                        cmd2 += ( ' -c "<< /PageSize [%s %s] >> setpagedevice" %s %s translate'
                                % (self.parse_bbox(self.input_bbox[0])) )
                    elif len(self.input_bbox[-1]) == 4:
                        # first is '(atend)' -> use last BoundingBox entry in the file
                        cmd2 += ( ' -c "<< /PageSize [%s %s] >> setpagedevice" %s %s translate'
                                % (self.parse_bbox(self.input_bbox[-1])) )
                elif len(self.input_page_bbox) > 0 and len(self.input_page_bbox[0]) == 4:
                    # try PageBoundingBox of first page
                    # won't work well for multi-page documents, maybe remove as option?
                    cmd2 += ( ' -c "<< /PageSize [%s %s] >> setpagedevice" %s %s translate'
                            % (self.parse_bbox(self.input_page_bbox[0])) )
            elif self.options.gs_dDEVICEPOINTS_opt:
                # custom paper size
                cmd1 += ( ' -dDEVICEWIDTHPOINTS=%s -dDEVICEHEIGHTPOINTS=%s'
                        % (self.options.gs_dDEVICEWIDTHPOINTS_val,
                            self.options.gs_dDEVICEHEIGHTPOINTS_val) )
            elif self.options.gs_sPAPERSIZE_opt:
                try:
                    # test for known PAPERSIZE
                    known_papersize = gs_papersizes[self.options.gs_sPAPERSIZE_val]
                    # orientation
                    if self.options.gs_sPAPERSIZE_orient == "portrait":
                        cmd1 += ( ' -dDEVICEWIDTHPOINTS=%s -dDEVICEHEIGHTPOINTS=%s'
                                % (known_papersize[0], known_papersize[1]) )
                    else: # Landscape
                        cmd1 += ( ' -dDEVICEWIDTHPOINTS=%s -dDEVICEHEIGHTPOINTS=%s'
                                % (known_papersize[1], known_papersize[0]) )
                except:
                    # gs_sPAPERSIZE_val not found in dict
                    # fall back to system default page
                    inkex.errormsg("PAPERSIZE not known to ghostscript, using system default.")
                    cmd1 += ''

            if self.options.gs_translate_xy_opt:
                if not self.options.gs_BOUNDINGBOX_opt:
                    cmd2 += ( ' -c %s %s translate' 
                            % (self.options.gs_translate_x_val,
                            self.options.gs_translate_y_val) )

            if self.options.gs_dFIXEDMEDIA_opt:
                cmd1 += ' -dFIXEDMEDIA'

            if self.options.gs_dORIENT1_swap:
                cmd1 += ' -dORIENT1=false'

            # EPS parameters
            if self.options.gs_dEPSbbox_opt == "crop":
                cmd1 += ' -dEPSCrop'
            elif self.options.gs_dEPSbbox_opt == "fit":
                cmd1 += ' -dEPSFitPage'

            if self.options.gs_dNOEPS_opt:
                cmd1 += ' -dNOEPS'

            ## run external command
            self.custom_import(infile, cmd0, cmd1, cmd2, cmd3)

        elif self.options.tab == '"unsupported_tab"':

            if (self.options.unsupported_cmd == "pstopdf"
                    and sys.platform == 'darwin'):
                cmd = '/usr/bin/pstopdf'
                if self.verbose:
                    inkex.debug(((cmd+' "%s" -o "%%s"') 
                        % infile.replace("%","%%"), "pstopdf", self.verbose))
                run((cmd+' "%s" -o "%%s"') 
                        % infile.replace("%","%%"), "pstopdf", self.verbose)
            else:
                # import with default settings
                inkex.errormsg("Not supported on %s, trying ps2pdf from Ghostscript instead." % sys.platform )
                self.default_import(infile)

        elif self.options.tab == '"help_tab"':
            # TODO: Provide help about available options

            # import with default settings
            self.default_import(infile)

        else: # unknown tab selected:
            # import with default settings
            self.default_import(infile)

if __name__ == '__main__':
    e = InputEPS()
    e.affect()

# vim: expandtab shiftwidth=4 tabstop=8 softtabstop=4 fileencoding=utf-8 textwidth=99
